"""
Our extensions to the loopy type system
"""
from dune.perftool.options import get_option
from dune.perftool.generation import function_mangler

import loopy as lp
import numpy as np


class VCLTypeRegistry:
    pass


def _populate_vcl_type_registry():
    VCLTypeRegistry.types = {}
    VCLTypeRegistry.names = {}

    # The base types that we are working with!
    for base_name, base_type, abbrev in [('float', np.float32, 'f'),
                                         ('double', np.float64, 'd'),
                                         ]:
        # The vector width in bits we are considering!
        for vector_bits in [128, 256, 512]:
            # Calculate the vector lane width
            count = vector_bits // (np.dtype(base_type).itemsize * 8)

            # Define the name of this vector type
            name = "Vec{}{}".format(count, abbrev)

            # Construct the numpy dtype!
            fieldnames = tuple(chr(ord("x") + i) for i in range(count))
            dtype = np.dtype(dict(names=fieldnames,
                                  formats=[base_type] * count,
                                  )
                             )

            VCLTypeRegistry.types[np.dtype(base_type), count] = dtype
            VCLTypeRegistry.names[dtype] = name


_populate_vcl_type_registry()


def get_vcl_type_size(nptype, register_size=None):
    if register_size is None:
        register_size = get_option("max_vector_width")

    return register_size // (np.dtype(nptype).itemsize * 8)


def get_vcl_type(nptype, register_size=None, vector_width=None):
    if vector_width is None:
        vector_width = get_vcl_type_size(nptype, register_size)

    return VCLTypeRegistry.types[np.dtype(nptype), vector_width]


def get_vcl_typename(nptype, register_size=None, vector_width=None):
    vcltype = get_vcl_type(nptype, register_size=register_size, vector_width=vector_width)
    return VCLTypeRegistry.names[vcltype]


class ExplicitVCLCast(lp.symbolic.FunctionIdentifier):
    def __init__(self, nptype, vector_width):
        self.nptype = nptype
        self.vector_width = vector_width

    def __getinitargs__(self):
        return (self.nptype, self.vector_width)

    @property
    def name(self):
        return get_vcl_typename(self.nptype, vector_width=self.vector_width)


@function_mangler
def vcl_cast_mangler(knl, func, arg_dtypes):
    if isinstance(func, ExplicitVCLCast):
        return lp.CallMangleInfo(func.name, (lp.types.NumpyType(func.nptype),), (arg_dtypes[0],))


@function_mangler
def vcl_function_mangler(knl, func, arg_dtypes):
    if func == "mul_add":
        vcl = lp.types.NumpyType(get_vcl_type(np.float64))
        return lp.CallMangleInfo("mul_add", (vcl,), (vcl, vcl, vcl))

    if func == "select":
        vcl = lp.types.NumpyType(get_vcl_type(np.float64))
        return lp.CallMangleInfo("select", (vcl,), (vcl, vcl, vcl))

    if func == "horizontal_add":
        vcl = lp.types.NumpyType(get_vcl_type(np.float64))
        return lp.CallMangleInfo("horizontal_add", (lp.types.NumpyType(np.float64),), (vcl,))
